import javax.sound.sampled.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import java.util.*;

/**
 * H4WriterExperiment - experiment software to test and demonstrate text input using Huffman coding.
 * <p>
 * 
 * If you are unfamiliar with <i>H4-Writer</i> begin by reading the first paper below.
 * <p>
 * 
 * <h3>Related references</h3>
 * <p>
 * 
 * The first paper below describes the rationale and specific details for designing text entry
 * methods using Huffman coding. It also details the design and evaluation <i>H4-Writer</i>. The
 * remaining papers describe extensions of <i>H4-Writer</i> to wearable computing, gesture input,
 * and smart phones, respectively.
 * <p>
 * 
 * <ul>
 * <li><a href="http://www.yorku.ca/mack/uist2011.html">1 thumb, 4 buttons, 20 words per minute:
 * Design and evaluation of H4-Writer</a>, by MacKenzie, Soukoreff, and Helga (<i>UIST 2011</i>)
 * <p>
 * 
 * <li><a href="http://www.yorku.ca/mack/iswc2012.html"> Huffman base-4 text entry glove
 * (H4-TEG)</a>, by Bajer, MacKenzie, and Baljko (<i>ISWC 2012</i>)
 * <p>
 * 
 * <li><a href="http://www.yorku.ca/mack/mhci2013e.html">Gestural text entry using Huffman
 * codes</a>, by Castellucci and MacKenzie (<i>MHCI 2013</i>)
 * <p>
 * <li><a href="http://www.yorku.ca/mack/ijvwhci2013.html">Eye on the message: Reducing attention
 * demand for touch-based text entry</a>, by MacKenzie and Castellucci (<i>Journal of Virtual Worlds
 * and Human-Computer Interaction</i>, 2013)
 * <p>
 * </ul>
 * 
 * <h3>Running the experiment software</h3>
 * <p>
 * 
 * <a href="http://www.yorku.ca/mack/HCIbook/Running/">Click here</a> for instructions on
 * launching/running the application.
 * <p>
 * 
 * <h3>Setup parameters</h3>
 * 
 * Upon launching, the application presents a setup dialog:
 * <p>
 * 
 * <center><img src = "H4WriterExperiment-1.jpg"></center>
 * <p>
 * 
 * The default parameter settings are read from a configuration file called
 * <code>H4WriterExperiment.cfg</code>. This file is created automatically when the application is
 * launched for the first time. The default parameter settings may be changed through the setup
 * dialog. The setup parameters are as follows:
 * <p>
 * 
 * <ul>
 * <li>Participant code - identifies the current participant. This is used in forming the names for
 * the output data files. Also, the sd2 output data file includes a column with the participant
 * code.
 * <p>
 * 
 * <li>Condition code - an arbitrary code used to associate a test condition with this invocation.
 * This parameter might be useful if the software is used in an experiment where a condition is not
 * inherently part of the application (e.g., Gender &rarr; male, female, Input Device &rarr;
 * gamepad, keyboard). The condition code is used in forming the name for the output data file.
 * Also, the sd2 output data file contains a column with the condition code.
 * <p>
 * 
 * <li>Session code - identifies the session. Useful if testing proceeds over multiple sessions to
 * gauge the progression of learning for <i>H4-Writer</i>.
 * <p>
 * 
 * Note: The setup dialog does not include an entry for "Block code". The block code is generated
 * automatically by the software.
 * <p>
 * 
 * <li>Number of phrases - specifies the number of phrases presented to the participant in the
 * current block.
 * <p>
 * 
 * <li>Phrases file - specifies a file containing phrases of text to be presented to participants
 * for entry. Phrases are drawn from the file at random. Typically, <a
 * href="phrases2.txt">phrases2.txt</a> is used. This is the phrase set published by MacKenzie and
 * Soukoreff in 2003 (<a href="http://www.yorku.ca/mack/chi03b.html">click here</a>). The other
 * files in the drop-down list are <a href="phrases2.txt">phrases2.txt</a>, <a
 * href="helloworld.txt">helloworld.txt</a>, <a href="helloworld2.txt">helloworld2.txt</a>, <a
 * href="quickbrownfox.txt">quickbrownfox.txt</a>, <a href="phrases.txt">phrases.txt</a>, <a
 * href="phrases100.txt">phrases100.txt</a>, <a href="vowells.txt">vowells.txt</a>, and <a
 * href="alphabet.txt">alphabet.txt</a>.
 * <p>
 * 
 * <li>Show presented text during input - a checkbox item that determines whether or not the
 * presented text is visible during entry of a phrase. Either way, the text phrase appears at the
 * beginning of a trial. If this option is unchecked, the phrase will disappear when the first
 * keystroke is entered.
 * <p>
 * 
 * <li>Lowercase only - a checkbox item that, if checked, forces all letters in the presented text
 * phrase to be lowercase.
 * <p>
 * 
 * <li>Show Huffman keyboard - a checkbox item that determines whether or not the Huffman keyboard
 * is visible in the UI. It might be useful to uncheck this option if input uses an external
 * keyboard and visual monitoring of the keys is to be avoided.
 * <p>
 * 
 * <li>Show symbols on keys - a checkbox item that determines whether or not symbols appear on the
 * keys. Uncheck this option if input uses the mouse and visual monitoring of the letters on keys is
 * to be avoided. This option is only relevant if the option above (Show Huffman keyboard) is
 * checked. Note, as well, that this option only applies to the basic code tree (letters, period,
 * comma, SPACE, BACKSPACE, ENTER, SYMBOL). If the user enters <code>[symbol]</code> to input a
 * punctuation or special character, rendering of the symbols on keys is restored during entry of
 * the symbol.
 * <p>
 * 
 * <li>Huffman code base - specifies the numeric base of the codes used in the Huffman code tree.
 * The possibilities are base 2 through to base 10.
 * <p>
 * 
 * <li>H4Writer gamepad layout - a checkbox item that determines whether to use a gamepad-style
 * layout for the keys. This option is only enabled if the option above (Huffman code base) is set
 * to 4. If the code base is 4 and this option is checked, the keys are laid out similar the
 * thumb-operated keys on the right side of a gamepad (see below).
 * <p>
 * </ul>
 * 
 * <h3>Program operation</h3>
 * 
 * The program can work with Huffman codes in base 2 through to base 10. The number of keys is equal
 * to the code base. The following are examples for base 2, base 4, and base 10: (click to enlarge)
 * <p>
 * 
 * <center><a href="H4WriterExperiment-2a.jpg"><img src = "H4WriterExperiment-2a.jpg"
 * width=600></a></center>
 * <p>
 * <center><a href="H4WriterExperiment-2b.jpg"><img src = "H4WriterExperiment-2b.jpg"
 * width=600></a></center>
 * <p>
 * <center><a href="H4WriterExperiment-2c.jpg"><img src = "H4WriterExperiment-2c.jpg"
 * width=800></a></center>
 * <p>
 * 
 * The virtual keys are laid out, left to right, as many as needed. The left-most key is selected by
 * pressing 1 on the system's keyboard. The next key is selected by pressing 2, and so on.
 * <p>
 * 
 * Mouse input is also supported. Just click on a key to select it.
 * <p>
 * 
 * There are two Huffman code trees for each code base, a basic tree containing the most common
 * symbols and codes (e.g., letters, period, comma, SPACE, BACKSPACE, ENTER, SYMBOL), and a symbol
 * tree for special symbols (e.g., digits, punctuation). Here are links to view the code trees:
 * <p>
 * 
 * <blockquote> Code base 2:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a
 * href="h2.txt">h2.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="h2-sym.txt">h2-sym.txt</a><br>
 * Code base 3:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a
 * href="h3.txt">h3.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="h3-sym.txt">h3-sym.txt</a><br>
 * Code base 4:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a
 * href="h4.txt">h4.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="h4-sym.txt">h4-sym.txt</a><br>
 * Code base 5:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a
 * href="h5.txt">h5.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="h5-sym.txt">h5-sym.txt</a><br>
 * Code base 6:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a
 * href="h6.txt">h6.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="h6-sym.txt">h6-sym.txt</a><br>
 * Code base 7:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a
 * href="h7.txt">h7.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="h7-sym.txt">h7-sym.txt</a><br>
 * Code base 8:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a
 * href="h8.txt">h8.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="h8-sym.txt">h8-sym.txt</a><br>
 * Code base 9:&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; <a
 * href="h9.txt">h9.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="h9-sym.txt">h9-sym.txt</a><br>
 * Code base 10:&nbsp;&nbsp;&nbsp;&nbsp;<a href="h10.txt">h10.txt<a/>&nbsp;&nbsp;&nbsp;&nbsp;<a
 * href="h10-sym.txt">h10-sym.txt</a> </blockquote>
 * <p>
 * 
 * In operation, entering "<code>[symbol]</code>" from the basic tree causes a transition to the
 * symbol tree. The next selection is from the symbol tree. If the next selection is "
 * <code>[SymLock]</code>" the symbol tree remains active for the input of multiple symbols, until "
 * <code>[SymLock]</code>" is again selected.
 * <p>
 * 
 * The base-4 case &mdash; <i>H4-Writer</i> &mdash; is a configuration of particular interest. The
 * GUI has a special look when the program is launched for 4-key input (if the option
 * "H4Writer gamepad layout" is checked):
 * <p>
 * 
 * <center><a href="H4WriterExperiment-3.jpg"><img src = "H4WriterExperiment-3.jpg"
 * width=600></a></center>
 * <p>
 * 
 * For this mode, it is intended that text will be entered with a four-key device, such as a
 * gamepad:
 * <p>
 * 
 * <center><img src="H4WriterExperiment-0.jpg"></center>
 * <p>
 * 
 * Input will use the D-pad on a gamepad, with LEFT = 1, UP = 2, RIGHT = 3, and DOWN = 4. For the
 * gamepad to work, software such as <a href="http://www.Xpadder.com/">Xpadder</a> must be installed
 * and configured so that the gamepad keys generate the appropriate keystroke events on the host
 * system. This is the mode of input used in the original paper presenting <i>H4-Writer</i> (<a
 * href="http://www.yorku.ca/mack/uist2011.html">click here</a> to view).
 * <p>
 * 
 * As established in the original test of <i>H4-Writer</i>, users learn the codes after a few hours
 * of practice. At this juncture, the rendering of the symbols on the keys is not needed. If the
 * setup parameter "Show symbols on keys" is unchecked, the symbols do not appear. Here's an example
 * for <i>H4-Writer</i>:
 * <p>
 * 
 * <center><a href="H4WriterExperiment-5.jpg"><img src = "H4WriterExperiment-5.jpg"
 * width=600></a></center>
 * <p>
 * 
 * 
 * <h3>Output data files</h3>
 * 
 * At the end of each block, data are written to two "summary data" (sd) files. The "sd1" file
 * contains data on a per-keystroke basis, one line per keystroke. The "sd2" file contains data on a
 * per-phrase basis, one line per phrase. Example data files follow:
 * 
 * <ul>
 * <li><a href="H4WriterExperiment-sd1-example.txt">sd1 example</a>
 * <li><a href="H4WriterExperiment-sd2-example.txt">sd2 example</a>
 * </ul>
 * <p>
 * 
 * The data in the sd2 file are full-precision, comma-delimited. Importing into a spreadsheet
 * application provides a convenient method to examine the data on a phrase-by-phrase basis. Here's
 * how the data might look after importing into Microsoft <i>Excel</i>: (click to enlarge)
 * <p>
 * 
 * <center><a href="H4WriterExperiment-4.jpg"><img src="H4WriterExperiment-4.jpg"
 * width=800></a></center>
 * <p>
 * 
 * Actual output files use "H4WriterExperiment" as the base filename. This is followed by the
 * participant code, the condition code, the session code, and the block code, for example,
 * <code>H4WriterExperiment-P01-C99-S07-B01.sd2</code>.
 * <p>
 * 
 * <h3>Miscellaneous</h3>
 * 
 * During entry, a phrase may be aborted by pressing "r" ("restart") on the system's keyboard. This
 * will terminate the current phrase and bring up a new phrase for input. "r" must be pressed before
 * "Enter" ("LU" for <i>H4-Writer</i>) is entered. Data for the aborted phrase are not saved.
 * <p>
 * 
 * If "r" is pressed between phrases (i.e., after "Enter" is entered), the effect is to erase the
 * newly-presented phrase and bring up a new phrase. Data for the previous phrase, which have
 * already been written to disk, are not affected.
 * <p>
 * 
 * When using this program in an experiment, it is a good idea to terminate all other applications
 * and disable the system's network connection. This will maintain the integrity of the data
 * collected and ensure that the program runs without hesitations.
 * <p>
 * 
 * @author Scott MacKenzie, 2010-2014
 */
public class H4WriterExperiment
{
	final static String APP_NAME = "H4WriterExperiment";
	final static int PARAMETERS = 11; // must equal the number of parameters defined below

	// setup parameters
	final static String[] PARTICIPANT_CODES = { "P00", "P01", "P02", "P03", "P04", "P05", "P06", "P07", "P08", "P09",
			"P10", "P11", "P12", "P13", "P14", "P15", "P16", "P17", "P18", "P19", "P20", "P21", "P22", "P23", "P24",
			"P25", "P26", "P27", "P28", "P29", "P30" };
	final static String[] CONDITION_CODE = { "C00" };
	final static String[] SESSION_CODES = { "S00", "S01", "S02", "S03", "S04", "S05", "S06", "S07", "S08", "S09",
			"S10", "S11", "S12", "S13", "S14", "S15", "S16", "S17", "S18", "S19", "S20", "S21", "S22", "S23", "S24",
			"S25", "S26", "S27", "S28", "S29", "S30" };
	final static String[] NUMBER_OF_PHRASES = { "5" };
	final static String[] PHRASES_FILE = { "phrases2.txt", "phrases.txt", "phrases100.txt", "helloworld.txt",
			"helloworld2.txt", "quickbrownfox.txt", "vowells.txt", "alphabet.txt" };
	final static String[] SHOW_PRESENTED = { "true" };
	final static String[] LOWERCASE_ONLY = { "true" };
	final static String[] SHOW_HUFFMAN_KEYBOARD = { "true" };
	final static String[] SHOW_SYMBOLS_ON_KEYS = { "true" };
	final static String[] HUFFMAN_CODE_BASE = { "4", "2", "3", "4", "5", "6", "7", "8", "9", "10" };
	final static String[] H4WRITER_GAMEPAD_LAYOUT = { "true" };

	// identify other files needed (used when execution from the .jar file for the 1st time)
	final static String[] OTHER_FILES = { "h2.txt", "h3.txt", "h4.txt", "h5.txt", "h6.txt", "h7.txt", "h8.txt",
			"h9.txt", "h10.txt", "h2-sym.txt", "h3-sym.txt", "h4-sym.txt", "h5-sym.txt", "h6-sym.txt", "h7-sym.txt",
			"h8-sym.txt", "h9-sym.txt", "h10-sym.txt", "phrases.txt", "phrases2.txt", "phrases100.txt",
			"helloworld.txt", "helloworld2.txt", "quickbrownfox.txt", "vowells.txt", "alphabet.txt", "click1.wav",
			"typewriter.wav", "miss.wav", "blip1.wav", "undo.wav" };

	public static void main(String[] args) throws IOException
	{
		// use Win32 look and feel
		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e)
		{
		}

		/*
		 * Define the default parameters settings. These are only used if the .cfg file does not
		 * exit, such as when the application is launched for the 1st time.
		 */
		String[] defaultParameters = new String[PARAMETERS];
		defaultParameters[0] = PARTICIPANT_CODES[0];
		defaultParameters[1] = CONDITION_CODE[0];
		defaultParameters[2] = SESSION_CODES[0];
		defaultParameters[3] = NUMBER_OF_PHRASES[0];
		defaultParameters[4] = PHRASES_FILE[0];
		defaultParameters[5] = SHOW_PRESENTED[0];
		defaultParameters[6] = LOWERCASE_ONLY[0];
		defaultParameters[7] = SHOW_HUFFMAN_KEYBOARD[0];
		defaultParameters[8] = SHOW_SYMBOLS_ON_KEYS[0];
		defaultParameters[9] = HUFFMAN_CODE_BASE[0];
		defaultParameters[10] = H4WRITER_GAMEPAD_LAYOUT[0];

		Configuration c = Configuration.readConfigurationData(APP_NAME, defaultParameters, OTHER_FILES);
		if (c == null)
		{
			System.out.println("Error reading configuration data from " + APP_NAME + ".cfg!");
			System.exit(0);
		}

		// -----------------------
		// create setup parameters
		// -----------------------

		SetupItemInfo[] sii = new SetupItemInfo[11];
		sii[0] = new SetupItemInfo(SetupItem.COMBO_BOX, "Participant code ", PARTICIPANT_CODES);
		sii[1] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Condition code ", CONDITION_CODE);
		sii[2] = new SetupItemInfo(SetupItem.COMBO_BOX, "Session code ", SESSION_CODES);
		sii[3] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Number of phrases ", NUMBER_OF_PHRASES);
		sii[4] = new SetupItemInfo(SetupItem.COMBO_BOX, "Phrases file  ", PHRASES_FILE);
		sii[5] = new SetupItemInfo(SetupItem.CHECK_BOX, "Show presented text during input ", SHOW_PRESENTED);
		sii[6] = new SetupItemInfo(SetupItem.CHECK_BOX, "Lowercase only ", LOWERCASE_ONLY);
		sii[7] = new SetupItemInfo(SetupItem.CHECK_BOX, "Show Huffman keyboard ", SHOW_HUFFMAN_KEYBOARD);
		sii[8] = new SetupItemInfo(SetupItem.CHECK_BOX, "Show symbols on keys ", SHOW_SYMBOLS_ON_KEYS);
		sii[9] = new SetupItemInfo(SetupItem.COMBO_BOX, "Huffman code base ", HUFFMAN_CODE_BASE);
		sii[10] = new SetupItemInfo(SetupItem.CHECK_BOX, "H4Writer gamepad layout ", H4WRITER_GAMEPAD_LAYOUT);

		// use setup to allow changes to default configuration
		Setup s = new Setup(null, c, APP_NAME, sii);
		s.showSetup(null);

		H4WriterExperimentGUI screen = new H4WriterExperimentGUI(c);
		screen.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		screen.setTitle("H4Writer Experiment");
		screen.pack();

		// put application in centre of screen
		int w = screen.getWidth();
		int h = screen.getHeight();
		Toolkit t = Toolkit.getDefaultToolkit();
		Dimension d = t.getScreenSize();
		screen.setLocation((d.width - w) / 2, (d.height - h) / 2);
		screen.setVisible(true);
	}
}

class H4WriterExperimentGUI extends JFrame implements KeyListener, MouseListener
{
	// the following avoids a "warning" with Java 1.5.0 complier (?)
	static final long serialVersionUID = 42L;

	final static String[] KEY_MAP = { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" };

	private BufferedWriter sd1File;
	private BufferedWriter sd2File;
	private Vector<Sample> samples;
	private String phrasesFile;
	private boolean showPresented;
	private boolean lowercaseOnly;
	private boolean showKeyboard;
	private boolean showSymbols;
	private boolean h4WriterSpecial;
	// private boolean newPhrase = true;
	private Random r = new Random();
	private JTextField presentedText;
	private JTextField transcribedText;

	private String[] phrases;
	private String pPhrase; // presented phrase
	private long t1, t2;
	private int count; // count keystrokes per phrase
	private int phraseCount;

	// new stuff here for huffman modifications
	private String[][] huffmanCodes; // the array actually used
	private String[][] huffmanLetterCodes;
	private String[][] huffmanSymbolCodes;
	private JTextArea[] huffmanKey;
	private JTextField keys;
	private String huffmanStem, huffmanStemPlusOne;
	private String huffmanLetterFile;
	private String huffmanSymbolFile;
	Clip tickClip;
	Clip typewriterClip;
	Clip missClip;
	Clip enterClip;
	Clip redoClip;
	boolean capsMode = false;
	boolean capsLockMode = false;
	boolean symbolMode = false;
	boolean symbolLockMode = false;

	private Font BIG = new Font("monospaced", Font.BOLD, 20);
	private Color BACKGROUND = new Color(254, 254, 218);
	private Color FOREGROUND = new Color(11, 11, 109);

	final String SPACE = "[Space]";
	final String BACKSPACE = "[Bksp]";
	final String ENTER = "[Enter]";
	final String CAPS = "[Cap]";
	final String CAPS_LOCK = "[CapsLock]";
	final String SYMBOL_MODE = "[Symbol]";
	final String SYMBOL_LOCK = "[SymLock]";

	Configuration c;
	String codeBase;
	// String[] keyMap = KEY_MAP;

	double[] blockWPM;
	double[] blockER;
	double[] blockEfficiency;
	int phraseIdx;
	JTextArea blockSummary;

	JDialog resultsDialog;
	JOptionPane resultsPane;
	JTextArea resultsArea;

	String participantCode, conditionCode, sessionCode, blockCode;
	String sd2Header;

	// constructor
	H4WriterExperimentGUI(Configuration cArg) throws IOException
	{
		c = cArg;
		participantCode = c.getConfigurationParameter(0);
		conditionCode = c.getConfigurationParameter(1);
		sessionCode = c.getConfigurationParameter(2);
		// blockCode = c.getBlockCode(); // blockCode computed automatically (see below)
		phraseCount = Integer.parseInt(c.getConfigurationParameter(3));
		phrasesFile = c.getConfigurationParameter(4);
		showPresented = Boolean.valueOf(c.getConfigurationParameter(5));
		lowercaseOnly = Boolean.valueOf(c.getConfigurationParameter(6));
		showKeyboard = Boolean.valueOf(c.getConfigurationParameter(7));
		showSymbols = Boolean.valueOf(c.getConfigurationParameter(8));
		codeBase = c.getConfigurationParameter(9);
		h4WriterSpecial = Boolean.valueOf(c.getConfigurationParameter(10));

		blockWPM = new double[phraseCount];
		blockER = new double[phraseCount];
		phraseIdx = 0;
		blockEfficiency = new double[phraseCount];
		// keyMap = KEY_MAP;

		sd2Header = "App,Participant,Condition,Session,Block,Min_keystrokes,Keystrokes,"
				+ "Presented_characters,Transcribed_characters,Time,MSD,WPM,ER,KSPC,Efficiency\n";

		// initialize phrases
		File f = new File(phrasesFile);
		if (!f.exists())
		{
			System.out.println("\nPhrases file not found - " + phrasesFile);
			System.exit(1);
		}

		// open disk file for input
		BufferedReader inFile = new BufferedReader(new FileReader(phrasesFile));
		Vector<String> v = new Vector<String>();
		String s;
		while ((s = inFile.readLine()) != null)
		{
			if (lowercaseOnly)
				s = s.toLowerCase();
			v.addElement(s);
		}

		phrases = new String[v.size()];
		v.copyInto(phrases);

		// initialize sound clips
		tickClip = initSound("click1.wav");
		typewriterClip = initSound("typewriter.wav");
		missClip = initSound("miss.wav");
		enterClip = initSound("blip1.wav");
		redoClip = initSound("undo.wav");

		// initialize Huffman stem (codes are added as keys are pressed)
		huffmanStem = "";
		huffmanStemPlusOne = "";

		// initialize Huffman letters + codes
		huffmanLetterFile = "h" + codeBase + ".txt";
		f = new File(huffmanLetterFile);
		if (!f.exists())
		{
			System.out.println("\nHuffmanletter code file not found - " + huffmanLetterFile);
			System.exit(1);
		}

		// initialize Huffman symbols + codes
		huffmanSymbolFile = "h" + codeBase + "-sym.txt";
		f = new File(huffmanSymbolFile);
		if (!f.exists())
		{
			System.out.println("\nHuffman symbol code file not found - " + huffmanSymbolFile);
			System.exit(1);
		}

		// open disk file for input (letters)
		inFile = new BufferedReader(new FileReader(huffmanLetterFile));
		Vector<String> v1 = new Vector<String>();
		Vector<String> v2 = new Vector<String>();
		while ((s = inFile.readLine()) != null && s.length() > 0)
		{
			StringTokenizer st = new StringTokenizer(s);
			if (st.countTokens() != 2)
			{
				System.out.println("Format error in Huffman letter code file: " + s);
				System.exit(0);
			}
			v1.addElement(st.nextToken()); // letter
			v2.addElement(st.nextToken()); // Huffman code
		}
		huffmanLetterCodes = new String[2][v1.size()];
		v1.copyInto(huffmanLetterCodes[0]);
		v2.copyInto(huffmanLetterCodes[1]);

		// open disk file for input (symbols)
		inFile = new BufferedReader(new FileReader(huffmanSymbolFile));
		v1 = new Vector<String>();
		v2 = new Vector<String>();
		while ((s = inFile.readLine()) != null && s.length() > 0)
		{
			StringTokenizer st = new StringTokenizer(s);
			if (st.countTokens() != 2)
			{
				System.out.println("Format error in Huffman symbol code file: " + s);
				System.exit(0);
			}
			v1.addElement(st.nextToken()); // symbol
			v2.addElement(st.nextToken()); // Huffman code
		}
		huffmanSymbolCodes = new String[2][v1.size()];
		v1.copyInto(huffmanSymbolCodes[0]);
		v2.copyInto(huffmanSymbolCodes[1]);

		huffmanCodes = huffmanLetterCodes; // default to letters

		// -----------------------
		// initialize output files
		// -----------------------

		blockCode = "";
		int blockCodeNumber = 0;
		String outputFile = "H4WriterExperiment-" + participantCode + "-" + conditionCode + "-" + sessionCode;
		String s1 = "";
		String s2 = "";
		do // find next available block code
		{
			++blockCodeNumber;
			blockCode = blockCodeNumber < 10 ? "B0" + blockCodeNumber : "B" + blockCodeNumber;
			s1 = outputFile + "-" + blockCode + ".sd1";
			s2 = outputFile + "-" + blockCode + ".sd2";
		} while (new File(s1).exists());

		sd1File = new BufferedWriter(new FileWriter(s1));
		sd2File = new BufferedWriter(new FileWriter(s2));

		// output header line to sd2 file
		try
		{
			sd2File.write(sd2Header, 0, sd2Header.length());
			sd2File.flush();
		} catch (IOException e)
		{
			System.err.println("ERROR WRITING HEADER LINE TO SD2 FILE!\n" + e);
			System.exit(1);
		}

		// initialize components

		resultsArea = new JTextArea(9, 40);
		resultsArea.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsArea.setBackground((new JButton()).getBackground());
		resultsPane = new JOptionPane(resultsArea, JOptionPane.INFORMATION_MESSAGE);
		resultsPane.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsDialog = resultsPane.createDialog(this, "Information");

		blockSummary = new JTextArea(5, 25);
		blockSummary.setFont(new Font("monospaced", Font.BOLD, 24));
		blockSummary.setBackground(new Color(255, 222, 222));
		blockSummary.setBorder(new EtchedBorder());

		presentedText = new JTextField(50);
		presentedText.setFocusable(false);
		presentedText.setFont(BIG);
		presentedText.setBackground(new Color(254, 254, 218));
		presentedText.setForeground(new Color(11, 11, 109));
		JLabel text1Label = new JLabel("  Presented:");
		text1Label.setFont(BIG);

		transcribedText = new JTextField(50);
		transcribedText.setFont(BIG);
		transcribedText.setBackground(BACKGROUND);
		transcribedText.setForeground(FOREGROUND);
		JLabel text2Label = new JLabel("Transcribed:");
		text2Label.setFont(BIG);
		transcribedText.setEditable(true);
		transcribedText.addKeyListener(this);

		keys = new JTextField(50);
		keys.setFocusable(false);
		keys.setFont(BIG);
		keys.setBackground(new Color(212, 208, 200));
		JLabel keysLabel = new JLabel("       Keys:");
		keysLabel.setFont(BIG);

		JPanel leftPanel = new JPanel(new GridLayout(0, 1, 0, 10));
		leftPanel.add(text1Label);
		leftPanel.add(text2Label);
		leftPanel.add(keysLabel);

		JPanel rightPanel = new JPanel(new GridLayout(0, 1, 0, 10));
		rightPanel.add(presentedText);
		rightPanel.add(transcribedText);
		rightPanel.add(keys);

		JPanel topPanel = new JPanel(new FlowLayout(FlowLayout.CENTER));
		topPanel.add(leftPanel, "West");
		topPanel.add(rightPanel, "East");

		// present first phrase for input
		pPhrase = phrases[r.nextInt(phrases.length)];
		presentedText.setText(pPhrase);
		transcribedText.setText("");
		t1 = 0;
		count = 0;
		samples = new Vector<Sample>();

		JPanel huffmanPanel = new JPanel();
		huffmanPanel.setBorder(new EtchedBorder());

		// special treatment for base-4 case (if H4Special selected)
		int codeBaseInt = Integer.parseInt(codeBase);
		if (codeBaseInt != 4 || (codeBaseInt == 4 && !h4WriterSpecial))
			huffmanPanel.setLayout(new GridLayout(1, codeBaseInt));
		else
			huffmanPanel.setLayout(new GridLayout(3, 3));

		LineBorder lb = new LineBorder(Color.red, 2);
		CompoundBorder cb = new CompoundBorder(new EtchedBorder(), lb);
		huffmanKey = new JTextArea[codeBaseInt];
		TitledBorder[] keyBorder = new TitledBorder[codeBaseInt];
		for (int i = 0; i < codeBaseInt; ++i)
		{
			huffmanKey[i] = new JTextArea(4, 15); // hmmm! size?
			huffmanKey[i].setFocusable(false);
			huffmanKey[i].setLineWrap(true);
			huffmanKey[i].setWrapStyleWord(true);
			huffmanKey[i].setFont(new Font("monospaced", Font.BOLD, 18));
			huffmanKey[i].setBackground(new Color(254, 254, 218));

			huffmanKey[i].addMouseListener(this);
			huffmanKey[i].setEditable(false);

			keyBorder[i] = new TitledBorder(cb);
			keyBorder[i].setTitleFont(BIG);
			huffmanKey[i].setBorder(keyBorder[i]);
			if (codeBaseInt != 4 || (codeBaseInt == 4 && !h4WriterSpecial))
				huffmanPanel.add(huffmanKey[i]);
		}

		if (codeBaseInt == 4 && h4WriterSpecial)
		{
			huffmanPanel.add(new JLabel(" "));
			huffmanPanel.add(huffmanKey[1]);
			huffmanPanel.add(new JLabel(" "));
			huffmanPanel.add(huffmanKey[0]);
			huffmanPanel.add(new JLabel(" "));
			huffmanPanel.add(huffmanKey[2]);
			huffmanPanel.add(new JLabel(" "));
			huffmanPanel.add(huffmanKey[3]);
			huffmanPanel.add(new JLabel(" "));
		}
		updateHuffmanKeys(huffmanStem); // initial stem is empty

		JPanel p = new JPanel();
		p.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
		p.setLayout(new BoxLayout(p, BoxLayout.Y_AXIS));
		p.add(topPanel);
		p.add(Box.createRigidArea(new Dimension(0, 10)));
		if (showKeyboard)
		{
			p.add(huffmanPanel);
			p.add(Box.createRigidArea(new Dimension(0, 10)));
		} else
			p.add(Box.createRigidArea(new Dimension(0, 50)));

		this.setContentPane(p);
	}

	private char getKeyCharFromMap(String keyTextArg)
	{
		boolean found = false;
		char c = '0';
		for (int i = 0; i < KEY_MAP.length; ++i)
		{
			if (KEY_MAP[i].equals(keyTextArg))
			{
				c = (char)('0' + i);
				found = true;
				break;
			}
		}
		if (!found)
		{
			// System.out.println("Error converting key text to char!");
			// System.exit(0);
		}
		return c;
	}

	private String getSymbolsForKey(int key, String stemArg)
	{
		String stem = stemArg + key;
		String s = "";
		for (int i = 0; i < huffmanCodes[0].length; ++i)
		{
			if (huffmanCodes[1][i].indexOf(stem) == 0)
				s += huffmanCodes[0][i] + " ";
		}
		return s.trim();
	}

	private void updateHuffmanKeys(String stem)
	{
		if (!showSymbols && !symbolMode)
			return;

		for (int i = 0; i < huffmanKey.length; ++i)
			huffmanKey[i].setText(getSymbolsForKey(i, stem));
	}

	private boolean isPrefix(String stem)
	{
		for (int i = 0; i < huffmanCodes[0].length; ++i)
		{
			if (huffmanCodes[1][i].equals(stem))
				return true;
		}
		return false;
	}

	private int sizeOfCode(String symbol, String[][] huffmanCodes)
	{
		int count = 0;
		for (int i = 0; i < huffmanCodes[1].length; ++i)
		{
			if (symbol.equals(huffmanCodes[0][i]))
			{
				count += huffmanCodes[1][i].length();
				break;
			}
		}
		if (count > 0)
			return count;
		else
		{
			System.out.println("Not found: " + symbol);
			return -1;
		}
	}

	private int minimumKeystrokes(String phrase)
	{
		int count = 0;
		for (int i = 0; i < phrase.length(); ++i)
		{
			Character c = new Character(phrase.charAt(i));
			String s = "" + c;

			// some pre-amble adjustments for SPACE, uppercase, and non-letters
			if (Character.isSpaceChar(c))
				s = SPACE; // re-map, if SPACE

			if (Character.isUpperCase(c)) // add transition to caps
			{
				count += sizeOfCode(CAPS, huffmanLetterCodes);
				s = s.toLowerCase();
			}

			if (Character.isLetter(c) || Character.isSpaceChar(c) || s.equals(".") || s.equals(","))
			{
				count += sizeOfCode(s, huffmanLetterCodes);
			} else
			{
				count += sizeOfCode(SYMBOL_MODE, huffmanLetterCodes);
				count += sizeOfCode(s, huffmanSymbolCodes);
			}
		}
		if (count > 0)
			return count;
		else
		{
			System.out.println("Oops!");
			return -1;
		}
	}

	private boolean valid(String stem)
	{
		for (int i = 0; i < huffmanCodes[0].length; ++i)
		{
			if (huffmanCodes[1][i].indexOf(stem) == 0)
				return true;
		}
		return false;
	}

	private String getHuffmanSymbol(String stem)
	{
		for (int i = 0; i < huffmanCodes[0].length; ++i)
		{
			if (huffmanCodes[1][i].equals(stem))
				return huffmanCodes[0][i];
		}
		return null;
	}

	// initialize stream for sound clip
	public Clip initSound(String soundFile)
	{
		AudioInputStream audioInputStream;
		Clip c = null;
		try
		{
			// Added for executable Jar:
			audioInputStream = AudioSystem.getAudioInputStream(new BufferedInputStream(getClass().getResourceAsStream(
					soundFile)));
			c = AudioSystem.getClip();
			c.open(audioInputStream);
		} catch (Exception e)
		{
			System.out.println("ERROR: Unable to load " + soundFile);
		}
		return c;
	}

	public void playSound(Clip c)
	{
		if (c != null)
		{
			c.setFramePosition(0); // rewind clip to beginning
			c.start(); // stops at end of clip
		}
	}

	// Key listener interface methods
	public void keyReleased(KeyEvent ke)
	{
	}

	public void keyPressed(KeyEvent ke)
	{
	}

	public void keyTyped(KeyEvent ke)
	{
		if (ke.getKeyChar() == 'r') // redo phrase
		{
			ke.consume();
			playSound(redoClip);
			prepareForNextPhrase();
			return;
		}

		// if (newPhrase)
		// {
		// newPhrase = false;
		// if (!showPresented)
		// presentedText.setText("");
		// }

		// remap jkl; to 0123 for home row input for H4Writer
		char temp = ke.getKeyChar();
		if (temp == 'j')
			temp = '0';
		if (temp == 'k')
			temp = '1';
		if (temp == 'l')
			temp = '2';
		if (temp == ';')
			temp = '3';

		char c = getKeyCharFromMap(KeyEvent.getKeyText(temp));
		--c; // remap 1 to 0, 2 to 1, etc.
		String key = "" + c;
		ke.consume(); // huffman keys don't go to transcribed text field
		processKey(key);
	}

	public void processKey(String keyArg)
	{
		String key = keyArg;
		String symbol = "";
		huffmanStem += key;
		huffmanStemPlusOne += "" + (char)(key.charAt(0) + 1);

		++count;
		if (t1 == 0)
		{
			t1 = System.currentTimeMillis();
			if (!showPresented)
				presentedText.setText("");
		}

		if (!valid(huffmanStem))
		{
			playSound(missClip);
			huffmanStem = huffmanStem.substring(0, huffmanStem.length() - 1);
			huffmanStemPlusOne = huffmanStemPlusOne.substring(0, huffmanStemPlusOne.length() - 1);
		} else if (isPrefix(huffmanStem))
		{
			symbol = getHuffmanSymbol(huffmanStem);

			// first, update the transcribe text field, as appropriate, and output
			// the appropriate audio
			String t = transcribedText.getText();
			if (symbol.equals(SPACE))
			{
				transcribedText.setText(t + " ");
				playSound(typewriterClip);
			} else if (symbol.equals(BACKSPACE))
			{
				if (t.length() >= 1)
				{
					transcribedText.setText(t.substring(0, t.length() - 1));
					playSound(typewriterClip);
				} else
					playSound(missClip);
			} else if (symbol.equals(CAPS))
			{
				capsMode = true;
				playSound(typewriterClip);
			} else if (symbol.equals(CAPS_LOCK))
			{
				capsLockMode = !capsLockMode;
				playSound(typewriterClip);
			} else if (symbol.equals(SYMBOL_MODE))
			{
				symbolMode = true;
				huffmanCodes = huffmanSymbolCodes;
				playSound(typewriterClip);
			} else if (symbol.equals(SYMBOL_LOCK))
			{
				symbolMode = false; // just in case last code was SYMBOL_MODE
				symbolLockMode = !symbolLockMode;
				if (symbolLockMode)
					huffmanCodes = huffmanSymbolCodes;
				else
					huffmanCodes = huffmanLetterCodes;
				playSound(typewriterClip);
			} else if (!symbol.equals(ENTER)) // add new character to phrase
			{
				if (symbolMode)
				{
					symbolMode = false;
					huffmanCodes = huffmanLetterCodes;
				}
				if (capsMode || capsLockMode)
				{
					symbol = symbol.toUpperCase();
					capsMode = false;
				}
				transcribedText.setText(t + symbol);
				playSound(typewriterClip);
			}

			huffmanStem = "";
			huffmanStemPlusOne = "";
			keys.setText(huffmanStemPlusOne);
			updateHuffmanKeys(huffmanStem);
		} else
		// not a prefix code
		{
			symbol = "";
			keys.setText(huffmanStemPlusOne);
			symbol = "";
			updateHuffmanKeys(huffmanStem);
			playSound(tickClip);
		}

		// collect data
		t2 = System.currentTimeMillis() - t1;
		samples.addElement(new Sample(t2, key, symbol));
		if (symbol.equals(ENTER))
		{
			playSound(enterClip);
			// newPhrase = true;

			// don't count the time or keystrokes for ENTER
			count = count - sizeOfCode(ENTER, huffmanLetterCodes);
			// backup from ENTER to get time (this assumes no error for ENTER)
			int backup = (sizeOfCode(ENTER, huffmanLetterCodes));
			t2 = samples.elementAt(samples.size() - (backup + 1)).time;

			String s1 = pPhrase;
			String s2 = transcribedText.getText();
			if (lowercaseOnly)
			{
				s1 = s1.toLowerCase();
				s2 = s2.toLowerCase();
			}

			MSD s1s2 = new MSD(s1, s2);
			int msd = s1s2.getMSD();

			int min = minimumKeystrokes(pPhrase);
			double kspc = (double)count / s2.length();
			double kspcmin = (double)min / s1.length();
			double efficiency = (kspcmin / kspc) * 100.0;

			// build formated results to display in GUI
			String resultsString = "Thank you!\n\n";
			resultsString += String.format("Presented text:\t%s\n", s1);
			resultsString += String.format("Transcribed text:\t%s\n", s2);
			resultsString += String.format("Entry speed:\t\t%.2f wpm\n", wpm(s2, t2));
			resultsString += String.format("Error rate:\t\t%.2f%%\n", s1s2.getErrorRateNew());
			resultsString += String.format("Efficiency:\t\t%.2f%%", efficiency);
			resultsArea.setText(resultsString);
			resultsDialog.setVisible(true);

			// build output data for sd2 file
			String sd2Stuff = "H4WriterExperiment," + participantCode + "," + conditionCode + "," + sessionCode + ","
					+ blockCode + ",";
			sd2Stuff += min + ","; // minimum number of keystrokes
			sd2Stuff += count + ","; // actual number of keystrokes
			sd2Stuff += s1.length() + ","; // presented characters
			sd2Stuff += s2.length() + ","; // number of characters produced
			sd2Stuff += String.format("%f,", t2 / 1000.0); // time (seconds)
			sd2Stuff += msd + ","; // minimum string distance
			sd2Stuff += String.format("%f,", wpm(s2, t2)); // entry speed (wpm)
			sd2Stuff += String.format("%f,", s1s2.getErrorRateNew()); // error rate (%)
			sd2Stuff += kspc + ","; // KSPC
			sd2Stuff += efficiency + "\n"; // efficiency (%)

			// build output data for sd1 file
			String sd1Stuff = pPhrase + "\n" + s2 + "\n";
			for (int i = 0; i < samples.size() - backup; ++i)
				sd1Stuff += samples.elementAt(i) + "\n";
			sd1Stuff += "#\n";

			// dump data
			try
			{
				sd1File.write(sd1Stuff, 0, sd1Stuff.length());
				sd1File.flush();
				sd2File.write(sd2Stuff, 0, sd2Stuff.length());
				sd2File.flush();

			} catch (IOException e)
			{
				System.err.println("ERROR WRITING TO DATA FILE!\n" + e);
				System.exit(1);
			}

			blockWPM[phraseIdx] = wpm(s2, t2);
			blockER[phraseIdx] = s1s2.getErrorRateNew();
			blockEfficiency[phraseIdx] = efficiency;
			++phraseIdx;

			// terminate if finished last phrase in block
			if (--phraseCount == 0)
			{
				String s = " Block summary\n";
				s += " =============\n";
				s += String.format(" Entry speed = %.1f wpm\n", mean(blockWPM));
				s += String.format(" Error rate  = %.1f%%\n", mean(blockER));
				s += String.format(" Efficiency  = %.1f%%", mean(blockEfficiency));

				blockSummary.setText(s);
				JOptionPane.showMessageDialog(this, blockSummary);
				try
				{
					sd1File.close();
					sd2File.close();
				} catch (IOException e)
				{
					System.err.println("ERROR CLOSING DATA FILES!\n" + e);
					System.exit(1);
				}
				System.exit(0);
			}

			prepareForNextPhrase();
		}
	}

	// calculate the mean of the values in a double array
	public static double mean(double n[])
	{
		double mean = 0.0;
		for (int j = 0; j < n.length; j++)
			mean += n[j];
		return mean / n.length;
	}

	void prepareForNextPhrase()
	{
		samples = new Vector<Sample>();
		pPhrase = phrases[r.nextInt(phrases.length)];
		presentedText.setText(pPhrase);
		transcribedText.setText("");
		t1 = 0;
		count = 0;
		capsLockMode = false;
		capsMode = false;
		return;
	}

	// compute typing speed in wpm given text entered and time in ms
	public static double wpm(String text, long msTime)
	{
		double speed = text.length();
		speed = speed / (msTime / 1000.0) * (60 / 5);
		return speed;
	}

	private class Sample
	{
		private long time;
		private String key;
		private String symbol;

		Sample(long timeArg, String keyArg, String symbolArg)
		{
			time = timeArg;
			key = keyArg;
			symbol = symbolArg;
		}

		@Override
		public String toString()
		{
			return time + " " + key + " " + symbol;
		}
	}

	@Override
	public void mouseClicked(MouseEvent me)
	{
	}

	@Override
	public void mouseEntered(MouseEvent me)
	{
	}

	@Override
	public void mouseExited(MouseEvent me)
	{
	}

	@Override
	public void mousePressed(MouseEvent me)
	{
	}

	@Override
	public void mouseReleased(MouseEvent me)
	{
		int i = -1;
		for (i = 0; i < huffmanKey.length; ++i)
		{
			if (me.getSource() == huffmanKey[i])
				break;
		}
		processKey("" + i);
	}
}